load.ConGenR<-function(){
  stringr.present<-installed.packages()
  packages<-stringr.present[,1]
  if(is.element("stringr", packages)==FALSE){
    install.packages('stringr')
    library('stringr')
  } else library('stringr')  
}

ConGen<-function(allele.calls, Class = NULL, rm.prefix = NULL, rm.vector = NULL, gte = TRUE, sex.id.markers = NULL, rm.sx.id = TRUE, PID = NULL){ 
  Sample.ID<-allele.calls$Sample.Name    
  allele.calls<-cbind(Sample.ID, allele.calls)
  
  temp<-rm.samples(rm.prefix, rm.vector, gen.data = allele.calls)
    allele.calls<-temp[[1]]
    samples.removed<-temp[[2]]

  temp<-add.class(Class, gen.data = allele.calls)
    allele.calls<-temp[[1]]
    duplicates.removed<-temp[[2]]

  allele.calls<-allele.calls[order(allele.calls$Class),] 
  allele.calls<-allele.calls[order(allele.calls$Marker, allele.calls$Sample.ID),]
  allele.calls[is.na(allele.calls)]<-0    
  alleles<-subset(allele.calls, select=c(Sample.ID, Class, Sample.File, Sample.Name, Marker, Allele.1,Allele.2,Allele.3,Allele.4))
  GenoType<-det.genotype(alleles)
  alleles<-cbind(alleles, GenoType)

  genotypes.complete.long<-alleles
  genotypes.complete.longna<-genotypes.complete.long
  is.na(genotypes.complete.longna[ , unlist(lapply(genotypes.complete.longna[,grep("^Allele", colnames(genotypes.complete.longna))], is.numeric))] ) <- 
    genotypes.complete.longna[ , unlist(lapply(genotypes.complete.longna[,grep("^Allele", colnames(genotypes.complete.longna))], is.numeric))] == 0

  temp<-det.congen(alleles)
    consensus.genotypes.long<-temp[[1]]
    consensus.genotypes.flat<-temp[[2]]

  temp<-long.to.flat.format(alleles=genotypes.complete.long)
    genotypes.complete.flat<-temp[[1]]
    genotypes.complete.flatna<-temp[[2]]
  
  if(gte==FALSE){
    temp<-pcr.success(gen.data = genotypes.complete.long)
      Overall.PCR.Success<-temp[[1]]
      PCR.Success.Rate<-temp[[2]]
      Locus.Rates<-temp[[3]]
      genotyping.errors<-NA
    temp<-sample.pcr.success(gen.data.long=genotypes.complete.long, gen.data.flat=genotypes.complete.flatna, PID = PID, sex.id.markers = sex.id.markers, rm.sx.id = rm.sx.id)
      Overall.Sample.PCR.Success<-temp[[1]]
      Sample.PCR.Success<-temp[[2]]
      Unique.Samples<-temp[[3]]
      Rates<-cbind(rbind(rbind(PCR.Success.Rate,Sample.PCR.Success),Unique.Samples),Overall = c(Overall.PCR.Success, Overall.Sample.PCR.Success, sum(Unique.Samples)))
      rownames(Rates)<-c("PCR.Success.Rate", "Sample.PCR.Success", "Unique.Samples")     
    } else {    
    temp<-gt.error(reference = consensus.genotypes.long, replicates = genotypes.complete.long)    
      genotyping.errors<-temp[[1]]
      Overall.PCR.Success<-temp[[2]]
      Class.Rates<-temp[[3]]
      Locus.Rates<-temp[[4]]
      Overall.ADO<-temp[[5]]
      Overall.FA<-temp[[6]]
    temp<-sample.pcr.success(gen.data.long=genotypes.complete.long, gen.data.flat=genotypes.complete.flat, PID = PID, sex.id.markers = sex.id.markers, rm.sx.id = rm.sx.id)
      Overall.Sample.PCR.Success<-temp[[1]]
      Sample.PCR.Success<-temp[[2]]
      Unique.Samples<-temp[[3]]      
      Rates<-cbind(rbind(rbind(Class.Rates,Sample.PCR.Success),Unique.Samples),Overall= c(Overall.FA, Overall.ADO, Overall.PCR.Success, Overall.Sample.PCR.Success, sum(Unique.Samples)))
      rownames(Rates)<-c("FA.Rate", "ADO.Rate", "PCR.Success.Rate", "Sample.PCR.Success", "Unique.Samples")     
    }
  temp<-samp.details(gen.data.long=genotypes.complete.long, con.gen.flat=consensus.genotypes.flat, sex.id.markers = sex.id.markers, rm.sx.id = rm.sx.id)
    Success.by.Number.of.Loci<-temp[[1]]
    Success.by.Sample<-temp[[2]]

  consensus.genotypes.long<-sort.alleles.long(consensus.genotypes.long)
  consensus.genotypes.flat<-sort.alleles.flat(consensus.genotypes.flat, nloci=length(levels(consensus.genotypes.long$Marker)), acols=min(grep(min(levels(consensus.genotypes.long$Marker)), colnames(consensus.genotypes.flat))), acole=max(grep(max(levels(consensus.genotypes.long$Marker)), colnames(consensus.genotypes.flat))))

  return(list(samples.removed = samples.removed, duplicates.removed = duplicates.removed, genotypes.complete.long = genotypes.complete.long,
          genotypes.complete.longna = genotypes.complete.longna, genotypes.complete.flat = genotypes.complete.flat, genotypes.complete.flatna = genotypes.complete.flatna, 
          consensus.genotypes.long = consensus.genotypes.long, consensus.genotypes.flat = consensus.genotypes.flat, genotyping.errors = genotyping.errors,
          Rates = Rates, Locus.Rates = Locus.Rates, Success.by.NLoci = Success.by.Number.of.Loci, Success.by.Sample = Success.by.Sample)) 
  }

gt.error<-function(reference, replicates){    
  header.names<-c("Sample.ID","Class","Sample.File","Sample.Name","Marker","Allele.1","Allele.2","Allele.3","Allele.4","GenoType")
  if(sum(names(reference[,1:10]) == names(replicates))==length(replicates[,1:10])){
    if(sum(header.names == names(replicates))==length(header.names)){
          if(length(levels(replicates$Sample.ID)[!(levels(replicates$Sample.ID) %in% levels(reference$Sample.ID))]) > 0 | 
            length(levels(reference$Sample.ID)[!(levels(reference$Sample.ID) %in% levels(replicates$Sample.ID))])>0){
              temp.x<-levels(replicates$Sample.ID)
              temp.y<-levels(reference$Sample.ID)
              rm.x<-temp.x[!(temp.x %in% temp.y)]
              rm.y<-temp.y[!(temp.y %in% temp.x)]
              rm.xy<-c(rm.x,rm.y)
              for(i in 1:length(rm.xy)){
                if(length(which(reference$Sample.ID==rm.xy[i]))>0){
                  reference<-reference[-which(reference$Sample.ID==rm.xy[i]),]
                  reference$Sample.ID<-factor(reference$Sample.ID)
                  }
                if(length(which(replicates$Sample.ID==rm.xy[i]))>0){
                  replicates<-replicates[-which(replicates$Sample.ID==rm.xy[i]),]
                  replicates$Sample.ID<-factor(replicates$Sample.ID)
                  }            
                }          
            }    
          temp<-combined.genotypes.long<-rbind(replicates[,1:10],reference[,1:10])  
          SA<-NULL; AA<-NULL; N.SA<-NULL; N.AA<-NULL; FA<-NULL; ADO<-NULL
          genotyping.errors<-temp[0,]
          for(i in 1:length(levels(temp$Sample.ID))){
            temp1<-subset(temp, temp$Sample.ID==levels(temp$Sample.ID)[i])    
            temp1$Marker<-factor(temp1$Marker)
            for(j in 1:length(levels(temp1$Marker))){ 
              temp2<-subset(temp1, temp1$Marker==levels(temp1$Marker)[j])
              temp2<-temp2[order(temp2$Class),] 
              temp3<-temp2[,c(grep("^Allele", colnames(temp2)),grep("^GenoType", colnames(temp2)))]
              temp.FA<-NULL; temp.ADO<-NULL; temp.SA<-NULL; temp.AA<-NULL; temp.N.SA<-NULL; temp.N.AA<-NULL
              if(temp3$GenoType[1]=="Incomplete" | temp3$GenoType[1]=="Sample_Failed" |temp3$GenoType[1]=="Requires_Review" | temp3$GenoType[1]==">2_Alleles"){    
                temp.FA<-NA
                temp.ADO<-NA
                temp.SA<-NA
                temp.AA<-NA
                temp.N.SA<-sum(temp3$Allele.1[2:length(temp3$Allele.1)]>0,na.rm=TRUE)
                temp.N.AA<-length(temp3$Allele.1[2:length(temp3$Allele.1)])
                for(p in 2:length(temp3$GenoType)){
                      temp.ADO<-append(temp.ADO,NA)
                      temp.FA<-append(temp.FA,NA)
                      temp.AA<-append(temp.AA,1)
                      temp.N.SA<-append(temp.N.SA,NA)
                      temp.N.AA<-append(temp.N.AA,NA)
                      if(temp3$Allele.1[p]==0 | is.na(temp3$Allele.1[p])){
                        temp.SA<-append(temp.SA,0)           
                      } else temp.SA<-append(temp.SA,1)
                    }
              } else if (temp3$GenoType[1]=="Heterozygote") {
                temp.FA<-NA
                temp.ADO<-NA
                temp.SA<-NA
                temp.AA<-NA
                temp.N.SA<-sum(temp3$Allele.1[2:length(temp3$Allele.1)]>0,na.rm=TRUE)
                temp.N.AA<-length(temp3$Allele.1[2:length(temp3$Allele.1)])
                is.na(temp3[-1 , unlist(lapply(temp3[-1,], is.numeric))])<-temp3[-1,unlist(lapply(temp3[-1,], is.numeric))]==0 
                for(v in 2:length(temp3$GenoType)){ 
                    ado<-0
                    fa<-0
                    if(is.na(temp3[v,1])){
                    ado<-NA
                    fa<-NA
                    } else {
                    for(k in 1:length(sort(as.numeric(temp3[1,1:4])))){ 
                      temp.av<-sort(as.numeric(temp3[1,1:4]))[k]  
                        if(any(sort(as.numeric(temp3[v,1:4]))==temp.av)){
                          ado<-ado
                          } else ado<-1
                    }            
                    for(n in 1:length(sort(as.numeric(temp3[v,1:4])))){ 
                      temp.rav<-sort(as.numeric(temp3[v,1:4]))
                      temp.av<-sort(as.numeric(temp3[1,1:4]))  
                        if(any(temp.av==temp.rav[n])){
                          fa<-fa
                          } else fa<-1
                     }
                    }
                    temp.ADO<-append(temp.ADO,ado)
                    temp.FA<-append(temp.FA,fa)
                    } 
                    for(p in 2:length(temp3$GenoType)){
                      temp.AA<-append(temp.AA,1)
                      temp.N.SA<-append(temp.N.SA,NA)
                      temp.N.AA<-append(temp.N.AA,NA)
                      if(temp3$Allele.1[p]<0 | is.na(temp3$Allele.1[p])){
                        temp.SA<-append(temp.SA,0)           
                      } else temp.SA<-append(temp.SA,1)
                    }
              } else if (temp3$GenoType[1]=="Homozygote") {
                temp.FA<-NA
                temp.ADO<-NA
                temp.SA<-NA
                temp.AA<-NA
                temp.N.SA<-sum(temp3$Allele.1[2:length(temp3$Allele.1)]>0,na.rm=TRUE)
                temp.N.AA<-length(temp3$Allele.1[2:length(temp3$Allele.1)])
                is.na(temp3[-1 , unlist(lapply(temp3[-1,], is.numeric))])<-temp3[-1,unlist(lapply(temp3[-1,], is.numeric))]==0 
                for(v in 2:length(temp3$GenoType)){ 
                    ado<-0
                    fa<-0
                    if(is.na(temp3[v,1])){
                    ado<-NA
                    fa<-NA
                    } else {
                    temp.av<-sort(as.numeric(temp3[1,1:4]))[1]  
                        if(any(sort(as.numeric(temp3[v,1:4]))==temp.av)){       
                          ado<-ado
                          } else ado<-1          
                    for(n in 1:length(sort(as.numeric(temp3[v,1:4])))){ 
                      temp.rav<-sort(as.numeric(temp3[v,1:4]))
                        if(any(temp.av==temp.rav[n])){  
                          fa<-fa
                          } else fa<-1
                     }
                    }
                    temp.ADO<-append(temp.ADO,ado)
                    temp.FA<-append(temp.FA,fa)
                    } 
                    for(p in 2:length(temp3$GenoType)){
                      temp.AA<-append(temp.AA,1)
                      temp.N.SA<-append(temp.N.SA,NA)
                      temp.N.AA<-append(temp.N.AA,NA)
                      if(temp3$Allele.1[p]<0 | is.na(temp3$Allele.1[p])){
                        temp.SA<-append(temp.SA,0)           
                      } else temp.SA<-append(temp.SA,1)
                    }
              } else if (temp3$GenoType[1]=="Homozygote_wUnc") {
                temp.FA<-NA
                temp.ADO<-NA
                temp.SA<-NA
                temp.AA<-NA
                temp.N.SA<-sum(temp3$Allele.1[2:length(temp3$Allele.1)]>0,na.rm=TRUE)
                temp.N.AA<-length(temp3$Allele.1[2:length(temp3$Allele.1)])
                is.na(temp3[-1 , unlist(lapply(temp3[-1,], is.numeric))])<-temp3[-1,unlist(lapply(temp3[-1,], is.numeric))]==0 
                for(v in 2:length(temp3$GenoType)){ 
                    ado<-0
                    fa<-0
                    if(is.na(temp3[v,1])){
                    ado<-NA
                    fa<-NA
                    } else { 
                    temp.av<-as.numeric(temp3[1,1])  
                        if(any(sort(as.numeric(temp3[v,1:4]))==temp.av)){
                          ado<-NA  
                          } else ado<-1         
                      temp.rav<-sort(as.numeric(temp3[v,1:4]))
                      if(length(temp.rav)==2){
                          if(any(temp.av==temp.rav)){
                           fa<-NA
                           } else fa<-1
                      } else if(length(temp.rav)>2){
                          fa<-1
                      } else if(length(temp.rav)<2){
                          fa<-NA
                        }
                    } 
                    temp.ADO<-append(temp.ADO,ado)
                    temp.FA<-append(temp.FA,fa)
                    } 
                    for(p in 2:length(temp3$GenoType)){
                      temp.AA<-append(temp.AA,1)
                      temp.N.SA<-append(temp.N.SA,NA)
                      temp.N.AA<-append(temp.N.AA,NA)
                      if(temp3$Allele.1[p]<0 | is.na(temp3$Allele.1[p])){
                        temp.SA<-append(temp.SA,0)           
                      } else temp.SA<-append(temp.SA,1)
                    }
              }
                SA<-append(SA,temp.SA)        
                AA<-append(AA,temp.AA)        
                N.SA<-append(N.SA, temp.N.SA) 
                N.AA<-append(N.AA, temp.N.AA)
                FA<-append(FA,temp.FA)
                ADO<-append(ADO,temp.ADO)
                genotyping.errors<-rbind(genotyping.errors,temp2)
            } 
          } 
          genotyping.errors<-cbind(genotyping.errors,FA)
          genotyping.errors<-cbind(genotyping.errors,ADO)
          genotyping.errors<-cbind(genotyping.errors,AA)
          genotyping.errors<-cbind(genotyping.errors,SA)
          genotyping.errors<-cbind(genotyping.errors,N.AA)
          genotyping.errors<-cbind(genotyping.errors,N.SA)
              
          temp.ge<-subset(genotyping.errors, genotyping.errors$Class!=0)     
          Rel.Class<-NULL   
          for(i in 1:length(temp.ge$Sample.ID)){
            if(temp.ge$Class[i]==999){
              Rel.Class<-append(Rel.Class,"N")
              } else Rel.Class<-append(Rel.Class, "Y")
          }
          temp.ge<-cbind(temp.ge,Rel.Class)
          temp.ge$Class<-as.factor(temp.ge$Class)   
          
          Class.PCR.Success<-100*(tapply(temp.ge$SA,temp.ge$Class,sum,na.rm=T)/tapply(temp.ge$AA,temp.ge$Class,sum,na.rm=T))
          Overall.PCR.Success<-100*(sum(temp.ge$SA)/sum(temp.ge$AA))
          
          temp.tpfa<-NULL; temp.tpado<-NULL; FA.rate.ra<-NULL; ADO.rate.ra<-NULL
          for(j in 1:length(levels(temp.ge$Class))){
            temp.tpfa<-append(temp.tpfa,(length(subset(temp.ge, temp.ge$Class==levels(temp.ge$Class)[j])$FA))-(sum(is.na(subset(temp.ge, temp.ge$Class==levels(temp.ge$Class)[j])$FA))))      
            temp.tpado<-append(temp.tpado, (length(subset(temp.ge, temp.ge$Class==levels(temp.ge$Class)[j])$ADO))-(sum(is.na(subset(temp.ge, temp.ge$Class==levels(temp.ge$Class)[j])$ADO)))) 
            }
          FA.rate.ra<-(as.numeric(tapply(temp.ge$FA,temp.ge$Class, sum,na.rm=T))/temp.tpfa)*100
          ADO.rate.ra<-(as.numeric(tapply(temp.ge$ADO,temp.ge$Class, sum,na.rm=T))/temp.tpado)*100  
          Class.ge.rates<-as.data.frame(rbind(FA.rate.ra,ADO.rate.ra))
          names(Class.ge.rates)<-sort(unique(temp.ge$Class))
          rownames(Class.ge.rates)<-c("FA.Rate", "ADO.Rate")
          Class.Rates<-rbind(Class.ge.rates, Class.PCR.Success = Class.PCR.Success) 
          
          Locus.PCR.Success<-100*(tapply(temp.ge$SA,temp.ge$Marker,sum,na.rm=T)/tapply(temp.ge$AA,temp.ge$Marker,sum,na.rm=T))
          temp.tpfa2<-NULL; temp.tpado2<-NULL; FA.rate.ra2<-NULL; ADO.rate.ra2<-NULL
          for(j in 1:length(levels(temp.ge$Marker))){
            temp.tpfa2<-append(temp.tpfa2,(length(subset(temp.ge, temp.ge$Marker==levels(temp.ge$Marker)[j])$FA))-(sum(is.na(subset(temp.ge, temp.ge$Marker==levels(temp.ge$Marker)[j])$FA))))      
            temp.tpado2<-append(temp.tpado2, (length(subset(temp.ge, temp.ge$Marker==levels(temp.ge$Marker)[j])$ADO))-(sum(is.na(subset(temp.ge, temp.ge$Marker==levels(temp.ge$Marker)[j])$ADO)))) 
            }
          FA.rate.ra2<-(as.numeric(tapply(temp.ge$FA,temp.ge$Marker, sum,na.rm=T))/temp.tpfa2)*100
          ADO.rate.ra2<-(as.numeric(tapply(temp.ge$ADO,temp.ge$Marker, sum,na.rm=T))/temp.tpado2)*100  
          Locus.Rates<-rbind(rbind(FA.rate.ra2,ADO.rate.ra2),Locus.PCR.Success)
          rownames(Locus.Rates)<-c("FA.Rate", "ADO.Rate", "Locus.PCR.Success")          
          Overall.ADO<-sum(tapply(temp.ge$ADO,temp.ge$Class, sum,na.rm=T))/sum(temp.tpado)*100
          Overall.FA<-sum(tapply(temp.ge$FA,temp.ge$Class, sum,na.rm=T))/sum(temp.tpfa)*100 
          
          genotyping.errors<-sort.alleles.long(genotyping.errors) 
      } else print ("columns do not match the required formatting")
    } else print("columns do not match the required formatting")
  return(list(genotyping.errors = genotyping.errors, Overall.PCR.Success = Overall.PCR.Success, Class.Rates = Class.Rates, Locus.Rates = Locus.Rates, Overall.ADO = Overall.ADO, Overall.FA = Overall.FA))
  }

pcr.success<-function(gen.data){
  Overall.PCR.Success<-sum(gen.data[,grep("^Allele", colnames(gen.data))[1]] > 0, na.rm=TRUE)/length(gen.data$Allele.1)*100
  Class.PCR.Success<-tapply(gen.data[,grep("^Allele", colnames(gen.data))[1]] > 0,gen.data$Class,sum,na.rm=T)/tapply(gen.data[,grep("^Allele", colnames(gen.data))[1]],gen.data$Class,length)*100
  Locus.PCR.Success<-tapply(gen.data[,grep("^Allele", colnames(gen.data))[1]] > 0,gen.data$Marker,sum,na.rm=T)/tapply(gen.data[,grep("^Allele", colnames(gen.data))[1]],gen.data$Marker,length)*100

  return(list(Overall.PCR.Success = Overall.PCR.Success, Class.PCR.Success = Class.PCR.Success, Locus.PCR.Success = Locus.PCR.Success))
  }

sample.pcr.success<-function(gen.data.long, gen.data.flat, PID = NULL, sex.id.markers = NULL, rm.sx.id = TRUE){
    if(sum(PID) > length(levels(gen.data.long$Marker))){
      print("Parameterization Error: Probabiility of identity (PID) value set to a value greater than the number of markers. PID cannot exceed the number of markers.")
      } else {
      if(is.null(PID)){        
        if(rm.sx.id == TRUE){  
          temp.nloci<-length(levels(gen.data.long$Marker))-length(sex.id.markers)   
          temp.suc.crit<-ceiling(temp.nloci/2)    
          } else {        
          temp.nloci<-length(levels(gen.data.long$Marker))  
          temp.suc.crit<-ceiling(temp.nloci/2)         
          }
        } else {        
        if(rm.sx.id == TRUE){
          temp.nloci<-length(levels(gen.data.long$Marker))-length(sex.id.markers)   
          temp.suc.crit<-PID    
          } else {
          temp.nloci<-length(levels(gen.data.long$Marker))  
          temp.suc.crit<-PID         
          }
        }     
        temp<-gen.data.flat
        temp.idx<-NULL
        if(rm.sx.id==TRUE && length(sex.id.markers>0)){
          for(i in 1:length(sex.id.markers)){
            temp.idx<-append(temp.idx,grep(paste(sex.id.markers[i],".",sep=""), colnames(temp), fixed=TRUE))
            }
          temp<-temp[,-c(temp.idx)]   
          } else temp<-temp    
          
        temp.allelelocs<-seq(from=grep(paste(levels(gen.data.long$Marker)[1],".", "1",sep=""), colnames(temp), fixed=TRUE), by=length(grep("^Allele", colnames(gen.data.long))),length.out=temp.nloci) 
        temp$Class<-as.factor(temp$Class)
        
        temp.Sample.ID<-NULL; temp.Sample.Class<-NULL; temp.succ<-NULL; temp.succ2<-NULL; temp.count<-NULL
        
        for (i in 1:length(levels(temp$Sample.ID))){ 
          temp1<-subset(temp, temp$Sample.ID==levels(temp$Sample.ID)[i])
          temp1$Class<-factor(temp1$Class)
          for (j in 1:length(levels(temp1$Class))){ 
            temp2<-subset(temp1, temp1$Class==levels(temp1$Class)[j])
            temp.Sample.ID<-append(temp.Sample.ID, as.character(temp2$Sample.ID[1]))
            temp.Sample.Class<-append(temp.Sample.Class, as.character(temp2$Class[1]))
              if(length(temp2$Sample.ID)<2){ 
                if(sum(sapply(temp2[1,temp.allelelocs],sum,na.rm=TRUE)>0)>=temp.suc.crit){ 
                  temp.succ<-append(temp.succ,1)
                  } else temp.succ<-append(temp.succ,0)   
                } else if (length(temp2$Sample.ID)>=2){ 
                  temp.row.count<-NULL
                  for(k in 1:nrow(temp2)){
                    temp.row.count<-append(temp.row.count,sum(sapply(temp2[k,temp.allelelocs],sum,na.rm=TRUE)>0))        
                  }         
                  temp.row.id<-which.max(temp.row.count)
                  if(sum(sapply(temp2[temp.row.id,temp.allelelocs],sum,na.rm=TRUE)>0)>=temp.suc.crit){           
                    temp.succ<-append(temp.succ,1)
                  } else temp.succ<-append(temp.succ,0) 
                } 
            } 
          } 
        temp.succ.df<-as.data.frame(cbind(cbind(temp.Sample.ID,temp.Sample.Class),temp.succ))
        temp.succ.df$temp.succ<-as.numeric(levels(temp.succ.df$temp.succ))[temp.succ.df$temp.succ]
        Sample.PCR.Success<-100*(tapply(temp.succ.df$temp.succ, temp.succ.df$temp.Sample.Class,sum)/tapply(temp.succ.df$temp.succ, temp.succ.df$temp.Sample.Class,length))
        Overall.Sample.PCR.Success<-100*sum((tapply(temp.succ.df$temp.succ, temp.succ.df$temp.Sample.Class,sum))/sum(tapply(temp.succ.df$temp.succ, temp.succ.df$temp.Sample.Class,length)))
        Unique.Samples<-tapply(temp.succ.df$temp.succ, temp.succ.df$temp.Sample.Class,length)       
    return(list(Overall.Sample.PCR.Success = Overall.Sample.PCR.Success, Sample.PCR.Success = Sample.PCR.Success, Unique.Samples = Unique.Samples))     
    }
  }

rm.samples<-function(rm.prefix = NULL, rm.vector = NULL, gen.data){
  idx<-NULL
  if(!is.null(rm.prefix)){
     if(!is.null(rm.vector)){
       RmSamples<-c(rm.prefix, as.character(rm.vector))
         } else RmSamples<-c(rm.prefix)
     } else {
         if(!is.null(rm.vector)){
           RmSamples<-c(as.character(rm.vector))
           } else {RmSamples<-NULL
             }
       }
  if(!is.null(RmSamples)){
    for(i in 1:length(RmSamples)){
      temp.idx<-grep(paste("^", RmSamples[i], sep=""),gen.data$Sample.ID)
      idx<-append(idx,temp.idx)
      }
      if(length(idx)>0){
        samples.removed<-gen.data[idx,]
        gen.data<-gen.data[-idx,]
        gen.data$Sample.ID<-factor(gen.data$Sample.ID)
        gen.data$Marker<-factor(gen.data$Marker)
        } else if (length(idx)==0){
          gen.data$Sample.ID<-factor(gen.data$Sample.ID)
          gen.data$Marker<-factor(gen.data$Marker)
          }
      } else {
        gen.data$Sample.ID<-factor(gen.data$Sample.ID)
        gen.data$Marker<-factor(gen.data$Marker)
        samples.removed<-NA
      }
  return(list(allele.calls = gen.data, samples.removed = samples.removed))
  }

add.class<-function(Class=NULL, gen.data){
  if(is.null(Class)){
    gen.data<-cbind(gen.data, Class=as.numeric(rep(999, length(gen.data[,1]))))
    duplicates.removed<-NA                           
  } else {
    gen.data$Sample.ID<-factor(gen.data$Sample.ID)  
    Class[,1]<-as.factor(Class[,1])
    Class[,2]<-as.numeric(Class[,2])
    temp.class<-Class[order(Class[,1]),]
    duplicates.removed<-temp.class[duplicated(temp.class[,1]),]   
    temp.class<-temp.class[!duplicated(temp.class[,1]),]                  
    temp.idx<-NULL
    temp.level<-levels(gen.data$Sample.ID)              
    temp.level2<-levels(temp.class[,1])
    for(i in 1:length(temp.class[,1])){                      
      if(any(temp.level==temp.level2[i])){                  
        temp.idx<-append(temp.idx, 1)                       
        } else temp.idx<-append(temp.idx,0)                 
      }   
    temp.class<-cbind(temp.class,temp.idx)                                
    temp.class<-subset(temp.class, temp.idx==1)                            
    temp.class$temp.idx<-NULL                                      
    colnames(temp.class) <- c("Sample.ID", "Class")
    temp.class$Sample.ID<-factor(temp.class$Sample.ID)                                
    temp.class$Class<-as.numeric(temp.class$Class)
    temp.class$Class[is.na(temp.class$Class)]<-(999)                          
    gen.data$Sample.Name<-factor(gen.data$Sample.Name) 
    gen.data$Sample.File<-factor(gen.data$Sample.File) 
    gen.data<-merge(gen.data, temp.class)                   
  }
  return(list(allele.calls = gen.data, duplicates.removed = duplicates.removed))
}  

det.congen<-function(alleles){
  consensus.genotypes.long<-alleles[0,]
  N.Alleles<-NULL; N.Reps<-NULL; N.Hetero<-NULL; N.Homo<-NULL; N.Failed<-NULL; N.Multiple<-NULL; FA<-NULL; ADO<-NULL
  for(i in 1:length(levels(alleles$Sample.ID))){
    temp1<-subset(alleles, alleles$Sample.ID==levels(alleles$Sample.ID)[i])
    temp1$Marker<-factor(temp1$Marker)
    for(j in 1:length(levels(temp1$Marker))){
      temp2<-subset(temp1, temp1$Marker==levels(temp1$Marker)[j])
      temp3<-temp2[,grep("^Allele", colnames(temp2))]
      for(k in 1:length(temp3[,1])){ 
      temp3[k,]<-sort(as.numeric(temp3[k,]))}
      temp4<-NULL
        m<-length(grep("^Allele", colnames(temp2)))
        for(m in 1:length(temp3[m,])){
          temp4<-append(temp4,as.numeric(temp3[,m]))}
          temp4<-as.data.frame(table(temp4))
          if(nrow(temp4)<2 & temp4[1,1]==0){
            cg<-c(NA,NA,NA,NA)            
            temp.alleles<-NULL  
            temp.counts<-NULL
            } else {
              temp4<-temp4[-as.numeric(temp4$temp4==0),]
              temp4$temp4<-factor(temp4$temp4)
              temp.alleles<-as.numeric(levels(temp4$temp4))[order(temp4$Freq,decreasing=TRUE)] 
              temp.counts<-temp4$Freq[order(temp4$Freq,decreasing=TRUE)]
              if(length(temp.alleles)==1){
                  if(temp.counts[1]>2) {cg<-c(rep(temp.alleles[1],2),rep(NA,(3-length(temp.alleles))))
                    } else if(temp.counts[1]==2) {cg<-c(temp.alleles[1],0,rep(NA,(3-length(temp.alleles))))
                    } else if(temp.counts[1]<2) {cg<-c(0,rep(NA,(4-length(temp.alleles))))
                  }    
                } else if(length(temp.alleles)==2) {
                  if(temp.counts[1]>1 & temp.counts[2]>1) {cg<-c(temp.alleles[1:2],rep(NA,(4-length(temp.alleles))))
                    } else if(temp.counts[1]==2 & temp.counts[2]<2) {cg<-c(temp.alleles[1],0,rep(NA,(4-length(temp.alleles))))
                    } else if(temp.counts[1]==3 & temp.counts[2]<2) {cg<-c(temp.alleles[1],0,rep(NA,(4-length(temp.alleles))))
                    } else if(temp.counts[1]>3 & temp.counts[2]<2) {cg<-c(rep(temp.alleles[1],2),rep(NA,(4-length(temp.alleles))))
                    } else if(temp.counts[1]<2) {cg<-c(0,0,rep(NA,(4-length(temp.alleles))))
                }
                } else if(length(temp.alleles)==3) {
                  if(temp.counts[1]>1 & temp.counts[2]>1 & temp.counts[3]>1) {cg<-c(temp.alleles[1:3],NA)
                    } else if(temp.counts[1]>1 & temp.counts[2]>1 & temp.counts[3]<2) {cg<-c(temp.alleles[1:2],NA,NA)
                    } else if(temp.counts[1]==2 & temp.counts[2]<2) {cg<-c(temp.alleles[1],0,NA,NA)
                    } else if(temp.counts[1]==3 & temp.counts[2]<2) {cg<-c(temp.alleles[1],0,NA,NA)
                    } else if(temp.counts[1]>3 & temp.counts[2]<2) {cg<-c(rep(temp.alleles[1],2),NA,NA)
                    } else if(temp.counts[1]<2) {cg<-c(rep(0,2),NA,NA)
                }
                } else if(length(temp.alleles)==4) {
                if(temp.counts[1]>1 & temp.counts[2]>1 & temp.counts[3]>1 & temp.counts[4]>1) {cg<-c(temp.alleles[1:4])
                  } else if(temp.counts[1]>1 & temp.counts[2]>1 & temp.counts[3]>1 & temp.counts[4]<2) {cg<-c(temp.alleles[1:3],NA)
                  } else if(temp.counts[1]>1 & temp.counts[2]>1 & temp.counts[3]<2) {cg<-c(temp.alleles[1:2],NA,NA)
                  } else if(temp.counts[1]==2 & temp.counts[2]<2) {cg<-c(temp.alleles[1],0,NA,NA)
                  } else if(temp.counts[1]==3 & temp.counts[2]<2) {cg<-c(temp.alleles[1],0,NA,NA)
                  } else if(temp.counts[1]>3 & temp.counts[2]<2) {cg<-c(rep(temp.alleles[1],2),NA,NA)
                  } else if(temp.counts[1]<2) {cg<-c(rep(0,2),NA,NA)
                }
                } else if(length(temp.alleles)>4) {
                if(temp.counts[1]>1 & temp.counts[2]>1 & temp.counts[3]>1 & temp.counts[4]>1) {cg<-c(temp.alleles[1:4])
                  } else if(temp.counts[1]>1 & temp.counts[2]>1 & temp.counts[3]>1 & temp.counts[4]<2) {cg<-c(temp.alleles[1:3],NA)
                  } else if(temp.counts[1]>1 & temp.counts[2]>1 & temp.counts[3]<2) {cg<-c(temp.alleles[1:2],NA,NA)
                  } else if(temp.counts[1]==2 & temp.counts[2]<2) {cg<-c(temp.alleles[1],0,NA,NA)
                  } else if(temp.counts[1]==3 & temp.counts[2]<2) {cg<-c(temp.alleles[1],0,NA,NA)
                  } else if(temp.counts[1]>3 & temp.counts[2]<2) {cg<-c(rep(temp.alleles[1],2),NA,NA)                               
                  } else if(temp.counts[1]<2) {cg<-rep(0,4)
                  }
                } 
              }
            temp.row<-temp2[1,]
            temp.row$Class<-0
            temp.row$Sample.File<-as.factor(paste(temp2$Sample.ID[1],".Con", sep=""))
            temp.row$Sample.Name<-as.factor(paste(temp2$Sample.ID[1],".Con", sep=""))
            temp.row[1,grep("^Allele", colnames(temp.row))]<-cg         
            consensus.genotypes.long<-rbind(consensus.genotypes.long, temp.row) 
            
            N.Alleles<-append(N.Alleles,length(temp.alleles))
            N.Reps<-append(N.Reps,length(temp2[,1]))
            N.Hetero<-append(N.Hetero,sum(temp2$GenoType=="Heterozygote"))
            N.Homo<-append(N.Homo,sum(temp2$GenoType=="Homozygote"))
            N.Failed<-append(N.Failed,sum(temp2$GenoType=="Sample_Failed"))
            N.Multiple<-append(N.Multiple,sum(temp2$GenoType==">2_Alleles"))  
            }
      }
  Allele1<-NULL; Allele2<-NULL; Allele3<-NULL; Allele4<-NULL; GenoType<-NULL
  for(i in 1:length(consensus.genotypes.long$Sample.ID)){
    if(is.na(consensus.genotypes.long[i,grep("^Allele", colnames(consensus.genotypes.long))[1]])){
      GenoType<-append(GenoType,"Sample_Failed")
    }
   else if(consensus.genotypes.long[i,grep("^Allele", colnames(consensus.genotypes.long))[1]]==0){
      GenoType<-append(GenoType,"Incomplete")
    } 
    else if(consensus.genotypes.long[i,grep("^Allele", colnames(consensus.genotypes.long))[1]]>0 & consensus.genotypes.long[i,grep("^Allele", colnames(consensus.genotypes.long))[1]]==consensus.genotypes.long[i,grep("^Allele", colnames(consensus.genotypes.long))[2]] 
      & (consensus.genotypes.long[i,grep("^Allele", colnames(consensus.genotypes.long))[3]]==0|is.na(consensus.genotypes.long[i,grep("^Allele", colnames(consensus.genotypes.long))[3]]))){
      GenoType<-append(GenoType,"Homozygote")
    } 
    else if(consensus.genotypes.long[i,grep("^Allele", colnames(consensus.genotypes.long))[1]]>0 & consensus.genotypes.long[i,grep("^Allele", colnames(consensus.genotypes.long))[2]]==0 & 
      is.na(consensus.genotypes.long[i,grep("^Allele", colnames(consensus.genotypes.long))[3]])){
      GenoType<-append(GenoType,"Homozygote_wUnc")
    } 
    else if(consensus.genotypes.long[i,grep("^Allele", colnames(consensus.genotypes.long))[1]]>0 & (consensus.genotypes.long[i,grep("^Allele", colnames(consensus.genotypes.long))[2]]>0 
      & consensus.genotypes.long[i,grep("^Allele", colnames(consensus.genotypes.long))[2]]!=consensus.genotypes.long[i,grep("^Allele", colnames(consensus.genotypes.long))[1]]) 
      & (consensus.genotypes.long[i,grep("^Allele", colnames(consensus.genotypes.long))[3]]==0|is.na(consensus.genotypes.long[i,grep("^Allele", colnames(consensus.genotypes.long))[3]]))){
      GenoType<-append(GenoType,"Heterozygote")
    } 
    else if(consensus.genotypes.long[i,grep("^Allele", colnames(consensus.genotypes.long))[1]]>0 & consensus.genotypes.long[i,grep("^Allele", colnames(consensus.genotypes.long))[2]]>0 
      & consensus.genotypes.long[i,grep("^Allele", colnames(consensus.genotypes.long))[3]]>0){ 
      GenoType<-append(GenoType,">2_Alleles")
    } 
    else {									
      GenoType<-append(GenoType,"Requires_Review")
    }
  }
  consensus.genotypes.long$GenoType<-as.factor(GenoType)
  consensus.genotypes.long<-cbind(consensus.genotypes.long, N.Alleles, N.Reps, N.Hetero, N.Homo, N.Failed, N.Multiple)  
  for(i in 1:length(levels(consensus.genotypes.long$Marker))){ 
    assign(paste("L",i,sep=""),subset(consensus.genotypes.long, consensus.genotypes.long$Marker==levels(consensus.genotypes.long$Marker)[i])) 
    }
  for(i in 1:length(levels(consensus.genotypes.long$Marker))){ 
    temp.df<-as.data.frame(get(paste("L",i,sep=""))) 
    for(j in 1:length(grep("^Allele", colnames(temp.df)))){
      if(j==1){
        j.int<-grep("^Allele", colnames(temp.df))[j]        
        names(temp.df)[j.int]<-paste(levels(temp.df$Marker)[i],".",j,sep="") 
        j.int<-j.int+1    
      } else{
        names(temp.df)[j.int]<-paste(levels(temp.df$Marker)[i],".",j,sep="")         
        j.int<-j.int+1    
        }
      }
    temp.df<-temp.df[,-which(names(temp.df) %in% "Marker")]               
    temp.df<-temp.df[,-which(names(temp.df) %in% "GenoType")]             
    temp.df<-temp.df[,-which(names(temp.df) %in% "N.Alleles")]            
    temp.df<-temp.df[,-which(names(temp.df) %in% "N.Reps")]               
    temp.df<-temp.df[,-which(names(temp.df) %in% "N.Hetero")]             
    temp.df<-temp.df[,-which(names(temp.df) %in% "N.Homo")]               
    temp.df<-temp.df[,-which(names(temp.df) %in% "N.Failed")]             
    temp.df<-temp.df[,-which(names(temp.df) %in% "N.Multiple")]           
    assign(paste("L",i,sep=""),temp.df)  
    }
  consensus.genotypes.flat<-L1
  for(i in 1:(length(levels(consensus.genotypes.long$Marker))-1)){ 
    consensus.genotypes.flat<-merge(consensus.genotypes.flat,as.data.frame(get(paste("L",i+1,sep=""))),all=TRUE) #Added add=TRUE argument following stephs data
    }
  return(list(consensus.genotypes.long = consensus.genotypes.long, consensus.genotypes.flat = consensus.genotypes.flat ))
  }

long.to.flat.format<-function(alleles){
  for(i in 1:length(levels(alleles$Marker))){
    assign(paste("L",i,sep=""),subset(alleles, alleles$Marker==levels(alleles$Marker)[i])) 
  }
  for(i in 1:length(levels(alleles$Marker))){ 
    temp.df<-as.data.frame(get(paste("L",i,sep=""))) 
    temp.y<-as.numeric(duplicated(temp.df$Sample.File)+1)           
    temp.df<-cbind(temp.df,temp.y)                                  
    temp.df$Sample.File<-paste(temp.df$Sample.File,temp.y,sep="-")  
    while(sum(duplicated(temp.df$Sample.File))>0){                  
        temp.df$temp.y<-(temp.df$temp.y + as.numeric(duplicated(temp.df$Sample.File)))    
        for(k in 1:length(temp.df$Sample.File)){                                          
          temp.rn<-str_locate(temp.df$Sample.File[k],"-")	                                
          temp.df$Sample.File[k]<-paste(substr(temp.df$Sample.File[k], 1,temp.rn),temp.df$temp.y[k],sep="")
          }                                                                               
    }                                                                                     
    temp.df$temp.y<-NULL          
    for(j in 1:length(grep("^Allele", colnames(temp.df)))){
      if(j==1){
        j.int<-grep("^Allele", colnames(temp.df))[j]
        names(temp.df)[j.int]<-paste(levels(temp.df$Marker)[i],".",j,sep="") 
        j.int<-j.int+1    
      } else{
        names(temp.df)[j.int]<-paste(levels(temp.df$Marker)[i],".",j,sep="") 
        j.int<-j.int+1    
        }
      }
    temp.df<-temp.df[,-which(names(temp.df) %in% "Marker")]               
    temp.df<-temp.df[,-which(names(temp.df) %in% "GenoType")]             
    assign(paste("L",i,sep=""),temp.df)  
  }
  genotypes.complete.flat<-L1
  for(i in 1:(length(levels(alleles$Marker))-1)){ 
    genotypes.complete.flat<-merge(genotypes.complete.flat,as.data.frame(get(paste("L",i+1,sep=""))),all=TRUE) #Added all=TRUE argument following Stephs data 
  }
  genotypes.complete.flatna<-genotypes.complete.flat
  is.na(genotypes.complete.flatna[ , unlist(lapply(genotypes.complete.flatna[,-c(1:4)], is.numeric))] ) <- 
    genotypes.complete.flatna[ , unlist(lapply(genotypes.complete.flatna[,-c(1:1:4)], is.numeric))] == 0
  return(list(genotypes.complete.flat = genotypes.complete.flat, genotypes.complete.flatna = genotypes.complete.flatna))
  }

samp.details<-function(gen.data.long, con.gen.flat, sex.id.markers = NULL, rm.sx.id = TRUE){
  if(rm.sx.id == TRUE){  
    temp.nloci<-length(levels(gen.data.long$Marker))-length(sex.id.markers) 
    } else {        
    temp.nloci<-length(levels(gen.data.long$Marker))  
    }
  temp<-con.gen.flat      
  temp.idx<-NULL
  if(rm.sx.id==TRUE && length(sex.id.markers>0)){
    for(i in 1:length(sex.id.markers)){
      temp.idx<-append(temp.idx,grep(paste(sex.id.markers[i],".",sep=""), colnames(temp), fixed=TRUE))
      }
    temp<-temp[,-c(temp.idx)] 
    } else temp<-temp
  temp.allelelocs1<-seq(from=grep(paste(levels(gen.data.long$Marker)[1],".", "1",sep=""), colnames(temp), fixed=TRUE), by=length(grep("^Allele", colnames(gen.data.long))),length.out=temp.nloci)
  temp.allelelocs2<-temp.allelelocs1+1
  temp$Class<-as.factor(temp$Class)
  temp.Sample.ID<-NULL
  temp.succ<-NULL
  for (i in 1:length(levels(temp$Sample.ID))){ 
    temp.Sample.ID<-append(temp.Sample.ID, as.character(temp$Sample.ID[i]))                               
    temp.succ<-append(temp.succ, sum((temp[i,temp.allelelocs1]>0 & temp[i,temp.allelelocs2]>0),na.rm=TRUE))  
    } 
  temp.succ.df<-as.data.frame(cbind(temp.Sample.ID,temp.succ))
  Sample.Succ<-as.data.frame(table(temp.succ.df$temp.succ))
  temp.succ.df$temp.succ<-as.numeric(as.character(temp.succ.df$temp.succ))
  temp.succ.df<-temp.succ.df[order(temp.succ.df$temp.succ),]
  colnames(Sample.Succ)<-c("Loci.w.CG","Number.of.Samples")
  colnames(temp.succ.df)<-c("Sample.ID","Loci.w.CG")
  return(list(Success.by.Number.of.Loci = Sample.Succ, Success.by.Sample = temp.succ.df))
  }

det.genotype<-function(alleles){
  Allele1<-NULL; Allele2<-NULL; Allele3<-NULL; Allele4<-NULL; GenoType<-NULL
  for(i in 1:length(alleles$Sample.ID)){	
    if(alleles[i,'Allele.1']==0){	
      Allele1<-append(Allele1,0)
      Allele2<-append(Allele2,0)
      Allele3<-append(Allele3,0)
      Allele4<-append(Allele4,0)
      GenoType<-append(GenoType,"Sample_Failed")
      }
    else if(alleles[i,'Allele.1']>0 & alleles[i,'Allele.2']==0){
      Allele1<-append(Allele1, alleles[i,'Allele.1'])
      Allele2<-append(Allele2, alleles[i,'Allele.1'])
      Allele3<-append(Allele3, alleles[i,'Allele.3'])
      Allele4<-append(Allele4, alleles[i,'Allele.4'])
      GenoType<-append(GenoType,"Homozygote")
      }
    else if(alleles[i,'Allele.1']>0 & alleles[i,'Allele.2']>0 & alleles[i,'Allele.3']==0){
      Allele1<-append(Allele1, alleles[i,'Allele.1'])
      Allele2<-append(Allele2, alleles[i,'Allele.2'])
      Allele3<-append(Allele3, alleles[i,'Allele.3'])
      Allele4<-append(Allele4, alleles[i,'Allele.4'])
      GenoType<-append(GenoType,"Heterozygote")
      }
    else {
      Allele1<-append(Allele1, alleles[i,'Allele.1'])
      Allele2<-append(Allele2, alleles[i,'Allele.2'])
      Allele3<-append(Allele3, alleles[i,'Allele.3'])
      Allele4<-append(Allele4, alleles[i,'Allele.4'])
      GenoType<-append(GenoType,">2_Alleles")
      }
    }
  return(GenoType)
  }

sort.alleles.long<-function(z){
  tempsort<-z
  temp.allele.cols<-grep("^Allele", colnames(tempsort))
  tempsort[,temp.allele.cols][tempsort[,temp.allele.cols]==0]<-9999
  for(i in 1:length(tempsort[,1])){ 
    if(all(is.na(tempsort[i,temp.allele.cols]))){
      tempsort[i,temp.allele.cols]<-NA
      } else {
        temp.str<-sort(tempsort[i,temp.allele.cols])
    if (length(temp.str>01)){
      tempsort[i,temp.allele.cols[1:length(temp.str)]]<-temp.str
      } else tempsort[i,temp.allele.cols]<-NA
    }
  }
  tempsort[,temp.allele.cols][tempsort[,temp.allele.cols]==9999]<-0
  z<-tempsort
  rm(list=ls(pattern="^temp"))
  return(z)
  }

sort.alleles.flat<-function(z, nloci, acols, acole){
  tempsort<-z
  temp.allele.cols<-acols
  temp.loci<-nloci
  temp.allele.cole<-acole
  tempsort[,temp.allele.cols:temp.allele.cole][tempsort[,temp.allele.cols:temp.allele.cole]==0]<-9999
  for(i in 1:length(tempsort[,1])){
    for(j in 1:temp.loci){
      k<-((j*4)+1)
      if(all(is.na(tempsort[i,k:(k+3)]))){
        tempsort[i,k:(k+3)]<-NA
        } else {
        temp.str<-sort(tempsort[i,k:(k+3)])
      if (length(temp.str>0)){
        tempsort[i,k:(k+length(temp.str)-1)]<-temp.str
      } else tempsort[i,k:(k+length(temp.str)-1)]<-NA
      }
    }
  }
  tempsort[,temp.allele.cols:temp.allele.cole][tempsort[,temp.allele.cols:temp.allele.cole]==9999]<-0
  z<-tempsort
  rm(list=ls(pattern="^temp"))
  return(z)
  }

flat.to.long<-function(gen.data.flat, loci.vec, reference = FALSE){
  temp.Sample.ID<-NULL; temp.Class<-NULL; temp.Sample.File<-NULL; temp.Sample.Name<-NULL; temp.Marker<-NULL
  temp.Allele.1<-NULL; temp.Allele.2<-NULL; temp.Allele.3<-NULL; temp.Allele.4<-NULL
  if(reference==FALSE){    
    for(i in 1:length(levels(gen.data.flat$Sample.ID))){
      temp1<-subset(gen.data.flat, gen.data.flat$Sample.ID==levels(gen.data.flat$Sample.ID)[i])
      for(j in 1:length(temp1$Sample.ID)){
        for(k in 1:length(loci.vec)){
          temp.Sample.ID<-append(temp.Sample.ID, as.character(temp1$Sample.ID[j]))
          temp.Class<-append(temp.Class, as.numeric(temp1$Class[j]))
          temp.Sample.File<-append(temp.Sample.File, as.character(temp1$Sample.File[j]))
          temp.Sample.Name<-append(temp.Sample.Name, as.character(temp1$Sample.Name[j]))
          temp.Marker<-append(temp.Marker, loci.vec[k])
          temp.Allele.1<-append(temp.Allele.1, temp1[j,grep(paste(loci.vec[k],".", 1,sep=""), colnames(temp1), fixed=TRUE)]) 
          temp.Allele.2<-append(temp.Allele.2, temp1[j,grep(paste(loci.vec[k],".", 2,sep=""), colnames(temp1), fixed=TRUE)])
          temp.Allele.3<-append(temp.Allele.3, temp1[j,grep(paste(loci.vec[k],".", 3,sep=""), colnames(temp1), fixed=TRUE)])
          temp.Allele.4<-append(temp.Allele.4, temp1[j,grep(paste(loci.vec[k],".", 4,sep=""), colnames(temp1), fixed=TRUE)])   
          }      
        }  
      }    
      temp.alleles<-data.frame(Allele.1 = temp.Allele.1, Allele.2 = temp.Allele.2, Allele.3 = temp.Allele.3, Allele.4 = temp.Allele.4)
      temp.gen.long<-cbind(Sample.ID = temp.Sample.ID, Class = temp.Class, Sample.File = temp.Sample.File, Sample.Name = temp.Sample.Name,Marker = temp.Marker, temp.alleles)
      temp.gt<-det.genotype(temp.gen.long)
      temp.gen.long<-cbind(temp.gen.long, GenoType = temp.gt)
    } else { 
    for(i in 1:length(levels(gen.data.flat$Sample.ID))){
      temp1<-subset(gen.data.flat, gen.data.flat$Sample.ID==levels(gen.data.flat$Sample.ID)[i])
        for(k in 1:length(loci.vec)){
          temp.Sample.ID<-append(temp.Sample.ID, as.character(temp1$Sample.ID[1]))
          temp.Class<-append(temp.Class, as.numeric(temp1$Class[1]))
          temp.Sample.File<-append(temp.Sample.File, as.character(temp1$Sample.File[1]))
          temp.Sample.Name<-append(temp.Sample.Name, as.character(temp1$Sample.Name[1]))
          temp.Marker<-append(temp.Marker, loci.vec[k])
          temp.Allele.1<-append(temp.Allele.1, temp1[1,grep(paste(loci.vec[k],".", 1,sep=""), colnames(temp1), fixed=TRUE)]) 
          temp.Allele.2<-append(temp.Allele.2, temp1[1,grep(paste(loci.vec[k],".", 2,sep=""), colnames(temp1), fixed=TRUE)])
          temp.Allele.3<-append(temp.Allele.3, temp1[1,grep(paste(loci.vec[k],".", 3,sep=""), colnames(temp1), fixed=TRUE)])
          temp.Allele.4<-append(temp.Allele.4, temp1[1,grep(paste(loci.vec[k],".", 4,sep=""), colnames(temp1), fixed=TRUE)])   
          }      
      }    
      temp.alleles<-data.frame(Allele.1 = temp.Allele.1, Allele.2 = temp.Allele.2, Allele.3 = temp.Allele.3, Allele.4 = temp.Allele.4)
      temp.gen.long<-cbind(Sample.ID = temp.Sample.ID, Class = temp.Class, Sample.File = temp.Sample.File, Sample.Name = temp.Sample.Name,Marker = temp.Marker, temp.alleles)
  
      Allele1<-NULL; Allele2<-NULL; Allele3<-NULL; Allele4<-NULL; GenoType<-NULL
      for(i in 1:length(temp.gen.long$Sample.ID)){
        if(is.na(temp.gen.long[i,grep("^Allele", colnames(temp.gen.long))[1]])){
          GenoType<-append(GenoType,"Sample_Failed")
        }
       else if(temp.gen.long[i,grep("^Allele", colnames(temp.gen.long))[1]]==0){
          GenoType<-append(GenoType,"Incomplete")
        } 
        else if(temp.gen.long[i,grep("^Allele", colnames(temp.gen.long))[1]]>0 & temp.gen.long[i,grep("^Allele", colnames(temp.gen.long))[1]]==temp.gen.long[i,grep("^Allele", colnames(temp.gen.long))[2]] 
          & (temp.gen.long[i,grep("^Allele", colnames(temp.gen.long))[3]]==0|is.na(temp.gen.long[i,grep("^Allele", colnames(temp.gen.long))[3]]))){
          GenoType<-append(GenoType,"Homozygote")
        } 
        else if(temp.gen.long[i,grep("^Allele", colnames(temp.gen.long))[1]]>0 & temp.gen.long[i,grep("^Allele", colnames(temp.gen.long))[2]]==0 & 
          is.na(temp.gen.long[i,grep("^Allele", colnames(temp.gen.long))[3]])){
          GenoType<-append(GenoType,"Homozygote_wUnc")
        } 
        else if(temp.gen.long[i,grep("^Allele", colnames(temp.gen.long))[1]]>0 & (temp.gen.long[i,grep("^Allele", colnames(temp.gen.long))[2]]>0 
          & temp.gen.long[i,grep("^Allele", colnames(temp.gen.long))[2]]!=temp.gen.long[i,grep("^Allele", colnames(temp.gen.long))[1]]) 
          & (temp.gen.long[i,grep("^Allele", colnames(temp.gen.long))[3]]==0|is.na(temp.gen.long[i,grep("^Allele", colnames(temp.gen.long))[3]]))){
          GenoType<-append(GenoType,"Heterozygote")
        } 
        else if(temp.gen.long[i,grep("^Allele", colnames(temp.gen.long))[1]]>0 & temp.gen.long[i,grep("^Allele", colnames(temp.gen.long))[2]]>0 
          & temp.gen.long[i,grep("^Allele", colnames(temp.gen.long))[3]]>0){ 
          GenoType<-append(GenoType,">2_Alleles")
        } 
        else {									
          GenoType<-append(GenoType,"Requires_Review")
        }
      }
      temp.gen.long<-cbind(temp.gen.long, GenoType)
   }  
  return(long.format.gen.data = temp.gen.long)
  } 

congen.matching<-function(congen.data.flat, congen.data.long = NULL, loci.names = NULL, matches.required = NULL, sex.id.markers = NULL,
                rm.sx.id = TRUE, consider.uncertainty = TRUE, skip.matches = TRUE, location.data = NULL, location.data.type = NULL){
    if(is.null(congen.data.long)){
      if(is.null(loci.names)){
        stop("If congen.data.long = NULL, user must provide a vector containing the names (as character strings) to the loci.names argument")
        } else {
        if(!is.null(sex.id.markers)){
          if(rm.sx.id==TRUE) loci.names<-loci.names[-c(match(sex.id.markers,loci.names))]
          }
        }
      } else {
      loci.names <-  levels(congen.data.long$Marker)
      if(!is.null(sex.id.markers)){
        if(rm.sx.id==TRUE) loci.names<-loci.names[-c(match(sex.id.markers,loci.names))]
        }
      }
    if(is.null(matches.required)) matches.required<-length(loci.names)
    if(matches.required > length(loci.names)) stop("matches.required cannot exceed the number of loci under consideration")

    loci.idx<-grep("Sample.ID", colnames(congen.data.flat))
    for(i in 1:length(loci.names)){
      loci.idx<-c(loci.idx,grep(paste("^", loci.names[i], sep=""), colnames(congen.data.flat))[1:2])
      }
    if(!is.null(location.data)){
      if(is.null(location.data.type)){
        stop("If location data is provided, location.data.type must be set to either 'Numeric' or 'Categorical'")
        }
    }
    cg.data<-congen.data.flat[,loci.idx]
    cg.data$Sample.ID<-factor(cg.data$Sample.ID)
    cg.data.orig<-cg.data
    matching.list<-list()
    if(consider.uncertainty==TRUE){
      matching.data<-cg.data[0,]
      new.col <- 1
      matching.data.template<-transform(matching.data, Mismatches=new.col[0])
      matching.data<-matching.data.template
      single.data<-matching.data.template
      for(focal.sample.i in 1:length(cg.data$Sample.ID)){
        temp.matching.data<-matching.data.template
        if(is.na(cg.data[focal.sample.i,]$Sample.ID)==FALSE){
          focal.sample <- cg.data[focal.sample.i,]
          focal.sample<-cbind(focal.sample, Mismatches=9999)
          cg.comp.data<-cg.data.orig[which(cg.data.orig$Sample.ID!=focal.sample$Sample.ID),]
          for(comp.sample.i in 1:length(cg.comp.data$Sample.ID)){
            comp.sample<-cg.comp.data[comp.sample.i,]
            n.matches<-0
            for(locus.i in 1:length(loci.names)){
              temp.call1<-focal.sample[grep(paste("^", loci.names[locus.i], sep=""), colnames(focal.sample))]
              temp.call2<-comp.sample[grep(paste("^", loci.names[locus.i], sep=""), colnames(comp.sample))]
              temp.csum<-c(match(temp.call1, temp.call2, nomatch = 0) >0, match(temp.call2, temp.call1, nomatch = 0) >0)
              temp.mm<-length(temp.csum)-sum(temp.csum)
              if(temp.mm == 0) n.matches<-n.matches+1
              }
            if(n.matches >= matches.required){
              comp.sample<-cbind(comp.sample, Mismatches=(length(loci.names)-n.matches))
              temp.matching.data<-rbind(temp.matching.data, comp.sample)
              if(skip.matches == TRUE){
                if(n.matches == length(loci.names)) cg.data$Sample.ID[cg.data$Sample.ID==comp.sample$Sample.ID]<-NA
                }
              }
            }
          if(length(temp.matching.data$Sample.ID)==0){
            single.data<-rbind(focal.sample, single.data)
            } else {
              temp.matching.data<-temp.matching.data[order(temp.matching.data$Mismatches, decreasing=FALSE),]
              temp.matching.data<-rbind(focal.sample, temp.matching.data)
              matches.present<-0
              if(length(matching.list)==0){
                matching.list[[1]]<- temp.matching.data
                } else {
                for(matching.set.i in 1:length(matching.list)) {
                sample.matches<-c(match(matching.list[[matching.set.i]]$Sample.ID, temp.matching.data$Sample.ID, nomatch = 0) >0,
                                  match(temp.matching.data$Sample.ID, matching.list[[matching.set.i]]$Sample.ID,nomatch = 0)  >0)
                if((length(sample.matches) - sum(sample.matches))==0) matches.present<-matches.present+1
                }
                if(matches.present==0) matching.list[[length(matching.list)+1]]<- temp.matching.data
                }
            }
          }
       }
       if(length(matching.list)>0){
          matching.list[[length(matching.list)+1]]<- single.data
          if(!is.null(location.data)){
            matching.list<-add.loc(list.data = matching.list, location.data = location.data, location.data.type=location.data.type)
            }
          matching.data<-do.call("rbind", matching.list)
          rownames(matching.data) <- 1:nrow(matching.data)
          }
    } else{
      matching.data<-cg.data[0,]
      new.col <- 1
      matching.data<-transform(matching.data, Mismatches=new.col[0])
      new.col <- 1
      matching.data.template<-transform(matching.data, Uncertain=new.col[0])
      matching.data<-matching.data.template
      single.data<-matching.data.template
      for(focal.sample.i in 1:length(cg.data$Sample.ID)){
        temp.matching.data<-matching.data.template
        if(is.na(cg.data[focal.sample.i,]$Sample.ID)==FALSE){
          focal.sample <- cg.data[focal.sample.i,]
          focal.sample<-cbind(focal.sample, Mismatches=9999)
          focal.sample<-cbind(focal.sample, Uncertain=NA)
          cg.comp.data<-cg.data.orig[which(cg.data.orig$Sample.ID!=focal.sample$Sample.ID),]
          for(comp.sample.i in 1:length(cg.comp.data$Sample.ID)){
            comp.sample<-cg.comp.data[comp.sample.i,]
            n.matches<-0
            n.uncertain<-0
            for(locus.i in 1:length(loci.names)){
              temp.call1<-focal.sample[grep(paste("^", loci.names[locus.i], sep=""), colnames(focal.sample))]
              temp.call2<-comp.sample[grep(paste("^", loci.names[locus.i], sep=""), colnames(comp.sample))]
              if(0 %in% c(temp.call1,temp.call2)==TRUE | any(is.na(c(temp.call1,temp.call2)))==TRUE){
                temp.csum<-c(match(temp.call1, temp.call2, nomatch = 0) >0, match(temp.call2, temp.call1, nomatch = 0) >0)
                temp.mm<-length(temp.csum)-sum(temp.csum)
                if(temp.mm == 0) {
                  n.matches<-n.matches+1
                  n.uncertain<-n.uncertain+1
                  } else { n.uncertain<-n.uncertain+1
                  }
                } else {
                  temp.csum<-c(temp.call1[1] %in% temp.call2, temp.call1[2] %in% temp.call2, temp.call2[1] %in% temp.call1, temp.call2[2] %in% temp.call1)
                  temp.mm<-length(temp.csum)-sum(temp.csum)
                  if(temp.mm == 0) n.matches<-n.matches+1
                  }
              }
            if(n.matches+n.uncertain >= matches.required){
              comp.sample<-cbind(comp.sample, Mismatches=(length(loci.names)-n.matches))
              comp.sample<-cbind(comp.sample, Uncertain=n.uncertain)
              temp.matching.data<-rbind(temp.matching.data, comp.sample)
              if(skip.matches == TRUE){
                if(n.matches == length(loci.names)) cg.data$Sample.ID[cg.data$Sample.ID==comp.sample$Sample.ID]<-NA
                }
              }
          }
          if(length(temp.matching.data$Sample.ID)==0){
            single.data<-rbind(focal.sample, single.data)
            } else {
              matches.present<-0
              if(length(matching.list)==0){
                temp.matching.data<-rbind(focal.sample, temp.matching.data)
                matching.list[[1]]<- temp.matching.data
                } else {
                temp.matching.data<-rbind(focal.sample, temp.matching.data)
                for(matching.set.i in 1:length(matching.list)) {
                sample.matches<-c(match(matching.list[[matching.set.i]]$Sample.ID, temp.matching.data$Sample.ID, nomatch = 0) >0,
                                  match(temp.matching.data$Sample.ID, matching.list[[matching.set.i]]$Sample.ID,nomatch = 0)  >0)
                if((length(sample.matches) - sum(sample.matches))==0) matches.present<-matches.present+1
                }
                if(matches.present==0) matching.list[[length(matching.list)+1]]<- temp.matching.data
                }
            }
          }
       }
       if(length(matching.list)>0){
          matching.list[[length(matching.list)+1]]<- single.data
          if(!is.null(location.data)){
            matching.list<-add.loc(list.data = matching.list, location.data = location.data, location.data.type=location.data.type)
            }
          matching.data<-do.call("rbind", matching.list)
          rownames(matching.data) <- 1:nrow(matching.data)
          }
    }
  return(list(matches.required = matches.required, consider.uncertainty = consider.uncertainty, skip.matches = skip.matches, matching.data = matching.data, matching.list = matching.list))
  }

add.loc<-function(list.data, location.data, location.data.type=c("Numeric", "Categorical")){
  if(location.data.type=="Categorical"){
    location.data[,1]<-factor(location.data[,1])
    location.data[,2]<-factor(location.data[,2])
    colnames(location.data) <- c("Sample.ID", "Location")
    for(list.i in 1:length(list.data)){
      temp.location<-NULL
      for(list.sample.i in 1:length(list.data[[list.i]]$Sample.ID)){
        temp.location<-append(temp.location, as.character(location.data$Location[which(as.character(location.data$Sample.ID)==as.character(list.data[[list.i]]$Sample.ID[list.sample.i]))]))
        }
      temp.location<-factor(temp.location)
      list.data[[list.i]]<-cbind(list.data[[list.i]], Location=temp.location)
      }
  } else if (location.data.type=="Numeric"){
    location.data[,1]<-factor(location.data[,1])
    location.data[,2]<-as.numeric(location.data[,2])
    location.data[,3]<-as.numeric(location.data[,3])
    colnames(location.data) <- c("Sample.ID", "X", "Y")
    for(list.i in 1:(length(list.data)-1)){
      temp.distance<-NULL
      temp.X1<-location.data$X[which(as.character(location.data$Sample.ID)==as.character(list.data[[list.i]]$Sample.ID[1]))]
      temp.Y1<-location.data$Y[which(as.character(location.data$Sample.ID)==as.character(list.data[[list.i]]$Sample.ID[1]))]
      for(list.sample.i in 1:length(list.data[[list.i]]$Sample.ID)){
        temp.X2<-location.data$X[which(as.character(location.data$Sample.ID)==as.character(list.data[[list.i]]$Sample.ID[list.sample.i]))]
        temp.Y2<-location.data$Y[which(as.character(location.data$Sample.ID)==as.character(list.data[[list.i]]$Sample.ID[list.sample.i]))]
        temp.distance<-append(temp.distance,sqrt((temp.X1-temp.X2)^2 + (temp.Y1-temp.Y2)^2))
        }
      temp.distance<-round(temp.distance,1)
      list.data[[list.i]]<-cbind(list.data[[list.i]], Distance=temp.distance)
      }
    list.data[[list.i+1]]<-cbind(list.data[[list.i+1]], Distance=rep(NA,length(list.data[[list.i+1]]$Sample.ID)))
  }
  return(matching.list = list.data)
}